/*
 * Copyright (C) 2009 by Eric Herman <eric@freesa.org>
 * Use and distribution licensed under the 
 * GNU Lesser General Public License (LGPL) version 2.1.
 * See the COPYING file in the parent directory for full text.
 */
package gearmanij;

import gearmanij.util.ByteUtils;
import gearmanij.util.IOUtil;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.channels.Channel;
import java.util.ArrayList;
import java.util.Collection;

/**
 * Requests and responses sent between workers or clients and job servers are
 * contained in binary packets. The packet header begins with a magic code that
 * indicates whether a packet is a request or a response. The packet also has a
 * type, which the receiver uses to determine what additional data should be in
 * the packet, as well as how to interpret the data. 
 */
public class Packet {

  private final PacketMagic magic;

  private final PacketType type;

  private final byte[] data;

  public Packet(PacketMagic magic, PacketType type, byte[] data) {
    this.magic = magic;
    this.type = type;
    this.data = ByteUtils.copy(data);
  }

  public Packet(InputStream in) {
    byte[] bytes = new byte[12];
    IOUtil.readFully(in, bytes);

    PacketHeader header = new PacketHeader(bytes);
    byte[] data = new byte[header.getDataLength()];
    if (data.length > 0) {
      IOUtil.readFully(in, data);
    }
    this.magic = header.getMagic();
    this.type = header.getType();
    this.data = data;
  }
  
  public Packet(ByteBuffer p) {
    byte [] headerbytes = new byte [Constants.GEARMAN_PACKET_HEADER_SIZE];
    byte [] data;
    
    int olimit = p.limit();
    p.rewind();
    p.limit(headerbytes.length);
    p.get(headerbytes);
    p.limit(olimit);
    PacketHeader header = new PacketHeader(headerbytes);
    
    data = new byte [p.remaining()];
    p.get(data);
    
    this.magic = header.getMagic();
    this.type = header.getType();
    this.data = data;
  }
  
  
  /**
   * @return a copy of the array
   */
  public byte[] getData() {
    return ByteUtils.copy(data);
  }

  /**
   * @return the length in bytes of the data
   */
  public int getDataSize() {
    return data.length;
  }

  public PacketType getPacketType() {
    return type;
  }

  public byte[] toBytes() {
    int totalSize = getDataSize() + 12;
    final ByteArrayOutputStream baos = new ByteArrayOutputStream(totalSize);
    write(baos);
    IOUtil.flush(baos);
    return baos.toByteArray();
  }
  
  public ByteBuffer toByteBuffer () {
    byte [] bytes = this.toBytes();
    ByteBuffer bb = ByteBuffer.allocate(bytes.length);
    bb.put(bytes);
    bb.rewind();
    return bb;
  }
  
  public ArrayList<byte []> getDataComponents(int tokens) {
    int curTokenStart, i, curToken;
    curTokenStart = i = curToken = 0;
    byte [] curTokenData = null;
    ArrayList <byte []> al = new ArrayList<byte []>(tokens);
    
    if (tokens == 0)
      return al;

    // process all but the last token (data segment)
    while  (curToken < tokens - 1 && i < data.length) {
      if (data[i] == ByteUtils.NULL) {
        curTokenData = new byte[i - curTokenStart];
        System.arraycopy(data, curTokenStart, curTokenData, 0, i - curTokenStart);
        al.add(curTokenData);
        curTokenStart = i + 1;
        curToken++;
      }
      i++;
    }
    
    // copy the last token
    curTokenData = new byte[data.length - i];
    System.arraycopy(data, curTokenStart, curTokenData, 0, data.length - i);
    al.add(curTokenData);
    return al;
  }
  

  /**
   * Writes the complete packet to the specified OutputStream.
   * @param os
   */
  public void write(OutputStream os) {
    /*
     * HEADER
     */
    new PacketHeader(magic, type, getDataSize()).write(os);

    /*
     * DATA
     * 
     * Arguments given in the data part are separated by a NULL byte, and the
     * last argument is determined by the size of data after the last NULL byte
     * separator. All job handle arguments must not be longer than 64 bytes,
     * including NULL terminator.
     */
    IOUtil.write(os, data);
  }

  public PacketType getType() {
    return type;
  }

  public String toString() {
    String s = magic + ":" + type + ":" + data.length;
    if (data.length > 0) {
      s += ": [" + ByteUtils.toHex(data) + "]";
    }
    return s;
  }

}
