package gearmanij.client;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;

import gearmanij.Constants;
import gearmanij.GearmanException;
import gearmanij.Packet;
import gearmanij.PacketMagic;
import gearmanij.PacketType;
import gearmanij.client.GearmanServerRequest.RequestState;

// TODO Logging
// TODO ThreadSafety?
// TODO change the server selection to use open connection

/**
 * <p>
 * The GearmanClient acts as the user-side interface for a Gearman System. The  
 * GearmanClient provides users with the ability to register Gearman JobServers 
 * with the client and then to submit jobs to the registered JobServers.
 * </p> 
 * <p>
 * The typical use pattern for a GearmanClient is as follows:
 * <ol>
 * <li>Create the client<br> <code>GearmanClient gc = new GearmanClient(); </code></li>
 * <li>Register a Gearman JobServer with client<br><code>gc.addServer("localhost");</code></li>
 * <li>Create a GearmanJob<br><code> String jobFunctionName = "reverse"; </code><br>
 * <code> byte [] jobData = jobFunctionName.getBytes(); </code><br>
 * <code> boolean detachJob = false;</code><br>
 * <code> GearmanJob myJob = new GearmanJob(jobFunctionName,jobData,detachJon);</code></li>
 * <li>Submit the job <br><code> gc.doJob(myJob);</code></li>
 * <li>Once an attached job has been submitted, users can either periodically poll 
 * the job to determine if it has completed<br><code>myJob.isCompleted();</code></li>
 * <li>Or the user can take an event based approach and handle all jobs have been 
 * updated.<br><code>Collection jobsWithActivity = gc.selectUpdatedJobs()</code></li>
 * <li>Once the user has finished with the client, the client should be closed.<br>
 * <code>gc.close()</code></li>
 * </ol>
 * The GearmanClient is not designed to be shared across threads. 
 * 
 */
public class GearmanClient {
  
  private static final String DESCRIPION_PREFIX = "GearmanClient";
  private final String DESCRIPTION;
  private HashMap<SelectionKey,GearmanJobServerSession> sessionsMap = null;
  private ArrayList<GearmanJobServerSession> sessionList = null;
  private Selector ioAvailable = null;
  private HashSet <GearmanJob> activeJobs = null;
  private HashSet <GearmanJob> updatedJobs = null;
  private Object listLock = new Object();
  private boolean isActive = true;
  
  /**
   * Create a new GearmanClient instance. Instances are not thread-safe and should 
   * not be shared across threads.
   */
  public GearmanClient () {
    sessionsMap = new HashMap<SelectionKey, GearmanJobServerSession>();
    sessionList = new ArrayList<GearmanJobServerSession>();
    activeJobs = new HashSet<GearmanJob>();
    updatedJobs = new LinkedHashSet<GearmanJob>();
    DESCRIPTION = new String(DESCRIPION_PREFIX + ":" + Thread.currentThread().getId());
  }
  
  /**
   * Is the client active.
   * 
   * @return True if the client is active and can accept new operations, else returns false. 
   */
  public boolean isActive() {
    return isActive;
  }
  
  /**
   * Registers a Gearman JobServer with the client. The location of the server 
   * is specified by the host parameter. This method assumes the server is listening
   * on the default port.
   * 
   * @param host The hostname or IP address of the host running the JobServer.
   * @throws IllegalArgumentException If an invalid hostname was specified
   * @throws IOException If an I/O error occurs.
   * @throws GearmanClient.ClientShutDown If client is shutdown 
   */
  public void addJobServer(String host) throws IllegalArgumentException, IOException, GearmanClient.ClientShutDown {
    addJobServer(host,Constants.GEARMAN_DEFAULT_TCP_PORT);
  }
  
  /**
   * Registers a Gearman JobServer with the client. The location of the server 
   * is specified by the host and port parameters. 
   * 
   * @param host The hostname or IP address of the host running the JobServer.
   * @param port The port number on which the JobServer is listening.
   * @throws IllegalArgumentException If an invalid hostname or port was specified
   * @throws IOException If an I/O error occurs. 
   * @throws GearmanClient.ClientShutDown If client is shutdown
   */
  public void addJobServer(String host, int port) throws IllegalArgumentException, IOException, GearmanClient.ClientShutDown {
    
    if (!isActive) {
      throw new GearmanClient.ClientShutDown();
    }
    
    InetSocketAddress sa = new InetSocketAddress(host,port);
    
    //this is a sub-optimal way to look for dups, but addJobServer
    //ops should be infrequent enough that this should be a big penalty
    Iterator<GearmanJobServerSession> iter = sessionList.iterator();
    while (iter.hasNext()) {
      if (iter.next().getServer().equals(sa)) {
        return;
      }
    }
    
    GearmanJobServerSession session = new GearmanJobServerSession(sa);
    sessionList.add(session);   

    if (ioAvailable == null) {
      ioAvailable = Selector.open();
    }

  }

  /**
   * Has a GearmanJobServer, listening on the default port, been registered with 
   * this client.
   * 
   * @param host The hostname or IP address of the host running the JobServer.
   * @return true if the job server has been registered witht client, else false.
   * @throws IllegalArgumentException  If an invalid hostname was specified
   * @throws GearmanClient.ClientShutDown If client is shutdown
   */
  public boolean hasJobServer (String host) throws IllegalArgumentException, GearmanClient.ClientShutDown {
    return hasJobServer(host,Constants.GEARMAN_DEFAULT_TCP_PORT);
  }
  
  /**
   * Has a GearmanJobServer been registered with this client.
   * 
   * @param host The hostname or IP address of the host running the JobServer.
   * @param port The port number on which the JobServer is listening.
   * @return true if the job server has been registered witht client, else false.
   * @throws IllegalArgumentException  If an invalid hostname or port was specified
   * @throws GearmanClient.ClientShutDown If client is shutdown
   */
  public boolean hasJobServer (String host, int port) throws IllegalArgumentException, GearmanClient.ClientShutDown {
    if (!isActive) {
      throw new GearmanClient.ClientShutDown();
    }
    return getSetOfJobServers().contains(new InetSocketAddress(host,port));
  }
  
  /**
   * Get the set of GearmanJobServers that have been registered with this client. 
   * 
   * @return A list of InetSocketAddresses for each JobServer registered with the system. 
   * @throws GearmanClient.ClientShutDown If the client is shutdown
   */
  public List<InetSocketAddress> getSetOfJobServers () throws GearmanClient.ClientShutDown {
    if (!isActive) {
      throw new GearmanClient.ClientShutDown();
    }
    
    ArrayList<InetSocketAddress> retSet = new ArrayList<InetSocketAddress>();
    Iterator<GearmanJobServerSession> iter = sessionList.iterator();
    while (iter.hasNext()) {
      retSet.add(iter.next().getServer());
    }
    return retSet;
  }
  
  /**
   * Submits a job to a Gearman JobServer. The server will then schedule the
   * job to be executed by a Gearman Worker. Performs initial validation of 
   * the job and then instructs the client to send the job to one of the 
   * Gearman JobServers. The method will return once the client has been asked
   * to send the job. 
   * 
   * @param job The job to be executed. 
   * @throws GearmanException If the job is invalid or an internal client 
   * error occurs.
   * @throws IOException If an I/O error occurs.
   */
  public void doJob(GearmanJob job) throws GearmanException, IOException {

    if (!isActive) {
      throw new GearmanClient.ClientShutDown();
    }
    
    if (job.isCompleted()) {
        return;
    }
        
    if ( activeJobs.contains(job)) {
      return;
    }
    Packet submitRequest = job.getRequest();
    job.setClient(this);
    GearmanServerRequest t = new GearmanServerRequest(job,submitRequest);
    synchronized (listLock) {
      activeJobs.add(job);
    }
    GearmanJobServerSession session = getSessionForTask();
    session.submitTaskToSession(t);
    
  }
  
  /**
   * Determines the state of a particular job. Typically this method is used to 
   * track the progress of detached jobs since a detached jobs provide no other 
   * means to track their status.
   * 
   * @param job The jobs who's state is being queried. 
   * @return A GearmanJobStatus object which reflects the state of the job when
   * the server received the updateJobStatusRequest. The GearmanJobStatus object
   * is a static snap-shot in time.
   * @throws IOException If an I/O error occurs.
   * @throws GearmanException If the job is invalid or an internal client 
   * error occurs.
   */
  public GearmanJobStatus updateJobStatus(GearmanJob job) throws IOException, GearmanException {
    return updateJobStatus(job.getHandle());
  }
  
  /**
   * Determines the state of a particular job. Typically this method is used to 
   * track the progress of detached jobs since a detached jobs provide no other 
   * means to track their status.
   * 
   * @param jobhandle The jobhandlde for the job who's state is being queried. 
   * @return A GearmanJobStatus object which reflects the state of the job when
   * the server received the updateJobStatusRequest. The GearmanJobStatus object
   * is a static snap-shot in time.
   * @throws IOException If an I/O error occurs.
   * @throws GearmanException If the job is invalid or an internal client 
   * error occurs.
   */
  public GearmanJobStatus updateJobStatus(byte [] jobhandle) throws IOException, GearmanException {
    if (!isActive) {
      throw new GearmanClient.ClientShutDown();
    }
    
    if (jobhandle == null || jobhandle.length == 0) {
      throw new GearmanException("Invalid job handle. Handle must not be null nor empty");
    }
    Packet statusRequest = new Packet(PacketMagic.REQ, PacketType.GET_STATUS,jobhandle);
    GearmanJobStatus handler = new GearmanJobStatus();
    GearmanServerRequest t = new GearmanServerRequest(handler,statusRequest);
    GearmanJobServerSession session = getSessionForTask();
    session.submitTaskToSession(t);
    while (!t.getState().equals(RequestState.FINISHED)) {
      driveClientIO();    
    }
    return handler; 
  }
  
  /**
   * Send an array of bytes to a Gearman JobServer and have the JobServer send the
   * array back to the client. 
   * 
   * @param data the array of bytes to be sent to the server.
   * @return an array of bytes which should be equal to the array passed into this 
   * method.
   * @throws IOException If an I/O error occurs.
   * @throws GearmanException If the job is invalid or an internal client 
   * error occurs.
   */
  public byte [] echo(byte [] data) throws IOException, GearmanException {
    if (!isActive) {
      throw new GearmanClient.ClientShutDown();
    }
    Packet echoRequest = new Packet(PacketMagic.REQ, PacketType.ECHO_REQ,data);
    GearmanEcho handler = new GearmanEcho();
    GearmanServerRequest t = new GearmanServerRequest(handler, echoRequest);
    GearmanJobServerSession session = getSessionForTask();
    session.submitTaskToSession(t);
    while (!t.getState().equals(RequestState.FINISHED)) {
      driveClientIO();    
    }
    return handler.getResults();
  }
  
  /**
   * Get the set of jobs which have been update since the last time this method
   * was executed. A job is said to be updated when it has received an event notification
   * from the job server. This could mean that job has completed or has received some 
   * transient event.
   * 
   * Users should take care to handle all Jobs returned by this method beforing invoking
   * it again. Consider the scenario where this method is invoked back-to-back. Assume that
   * that prior to the first invocation, there exists 10 jobs which have been updated and 
   * that no events occur between the first and second invocation. In this scenario the  
   * first invocation will return a collection of the 10 jobs. The second invocation, which 
   * immediately follows the first, will result an empty collection. So user must be sure they
   * have handled all events returned by prior invocations or at least have a handle to the 
   * prior collection before invoking this method.
   * 
   * @return The set of jobs which received an update notification since the prior 
   * execution of this method.  
   * @throws IOException IOException If an I/O error occurs.
   * @throws GearmanException If the job is invalid or an internal client 
   * error occurs.
   */
  public Collection <GearmanJob> selectUpdatedJobs () throws IOException,GearmanException {
    if (!isActive) {
      throw new GearmanClient.ClientShutDown();
    }
    Collection <GearmanJob>retSet = new HashSet<GearmanJob>();
    driveClientIO();
    synchronized (listLock) {
      if (!updatedJobs.isEmpty()) {
        retSet = (Collection <GearmanJob>)updatedJobs.clone();
        updatedJobs.clear();
      }
    }
    return retSet;
  }
  
  /**
   * Shutdown the client. This method closes all connections between the client and
   * the job servers that have been registered with the client. A client that has been 
   * shutdown may not be re-used and any attempt to invoke a method on the client 
   * will result in an Exception.
   * 
   * @throws IOException If an I/O error occurs while terminating communication 
   * with the GearmanJobServers. 
   */
  public void shutdown() throws IOException {
    isActive = false;
    StringBuffer hostCloseFailed = new StringBuffer();
    boolean error = false;
    Iterator<GearmanJobServerSession> sessions = sessionList.iterator();
    while (sessions.hasNext()) {
      GearmanJobServerSession curSession = sessions.next();
      try {
        curSession.closeSession();
      } catch (Exception e) {
        error = true;
        hostCloseFailed.append(curSession + "failed to close: " + e.getMessage() + System.getProperty("line.separator"));
      }
      sessions.remove();
    }
    sessionsMap.clear();
    sessionsMap = null;
    sessionList.clear();
    sessionList = null;    
    activeJobs.clear();
    activeJobs = null;
    updatedJobs.clear();
    updatedJobs = null;
    if (error)
      throw new IOException("The following exceptions were encounterd while closing the client:" + 
          System.getProperty("line.separator") + hostCloseFailed.toString());
  }
  
  /**
   * Retrieve the number of active jobs currently being managed by this client. An active job
   * is a job that has been submitted to the client but has not been completed.
   * 
   * @return The number of active jobs.
   * 
   * @throws GearmanClient.ClientShutDown If client is shutdown.
   */
  public int getNumberofActiveJobs() throws GearmanClient.ClientShutDown {
    if (!isActive) {
      throw new GearmanClient.ClientShutDown();
    }
    return activeJobs.size();
  }
  
  public String toString() {
    return DESCRIPTION;
  }
  
  void handleJobUpdate(GearmanJob job) throws GearmanException {
    if (!isActive) {
      throw new GearmanClient.ClientShutDown();
    }
    synchronized (listLock) {
      updatedJobs.add(job);
      if (job.isCompleted()) {
        activeJobs.remove(job); 
      }
    } 
  }
  
  private GearmanJobServerSession getSessionForTask() throws IOException {
    if (sessionList.isEmpty()) {
      throw new IOException("No servers registered with client");
    } 
    
    int s = (int)Math.round((Math.random() * (sessionList.size() -1)));
    GearmanJobServerSession session = sessionList.get(s);
    if (!session.isInitialized()) {
      SocketChannel sc = session.initSession();
      SelectionKey key = sc.register(ioAvailable, SelectionKey.OP_READ | SelectionKey.OP_WRITE);
      sessionsMap.put(key, session);
    }    
    return session;
  }
  
  private void driveClientIO() throws IOException, GearmanException {   
      ioAvailable.selectNow();
      Iterator<SelectionKey> iter = ioAvailable.selectedKeys().iterator();
      while (iter.hasNext()) {
        SelectionKey key = iter.next();
        GearmanJobServerSession s = sessionsMap.get(key);
        s.driveSessionIO();
      }
  } 
  
  public class ClientShutDown extends GearmanException {
    
  }
  
}


