package gearmanij.client;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.UUID;

import gearmanij.Constants;
import gearmanij.GearmanException;
import gearmanij.Packet;
import gearmanij.PacketMagic;
import gearmanij.PacketType;
import gearmanij.util.ByteUtils;

//TODO The UUID implementation could be improved

/**
 * <p>
 * A GearmanJob is a basic unit of work handled by a Gearman
 * system. A GearmanJob is created by a user of a Gearman system. 
 * The user than passes the GearmanJob to a instance of a {@link GearmanClient}
 * which then submits the job to GearmanJobServer. The GearmanJobServer
 * oversees the scheduling and execution of the GearmanJob and is responsible
 * for managing communication between the GearmanWorker, which is executing 
 * the job, and the GearmanClient and GearmanJob handle.
 * </p>
 * <p>A GearmanJob has the following attribute
 * <ul>
 * <li><b>Funciontion Name:</b> The name of the worker function to be executed by the GearmanWorker</li>
 * <li><b>Data:</b> The data passed to function executed by the GearmanWorker.</li>
 * <li><b>Classification:</b> The classification of this job instance, either attached or detached (see below).</li>
 * <li><b>Priority:</b> The priority to which this job is given</li> 
 * <ul>
 * </p> 
 * <p>
 * There are two classifications of GearmanJobs within the Gearman System
 * <ol>
 * <li><b>attached jobs:</b>An attached job can receive updates and notifications 
 * from the GearmanWorker. Typically, as a job progress through its execution cycle,
 * the GearmanWorker and GearmanJobServer will send out notifications  as certain
 * milestones are completed. Example of such notifications include, the submission
 * the job to a worker, the completion of the job,  or if job has attempted to send 
 * data back to the client. </li>
 * <li><b>detached jobs</b>A detached job is a job that is executed in a 'fire-and-forget'
 * mode. Detached jobs will not receive any notifications or events that have been 
 * generated by the GearmanWorker. To learn the state of a detached job, the user must
 * explicitly ask the GearmanClient to retrieve the state of the job. This is accomplished
 * via the {@link GearmanClient#updateJobStatus(GearmanJob)} call.</li>
 * </ol>
 * </p>
 * 
 */
public class GearmanJob implements GearmanServerResponseHandler {


  public static enum JobPriority {
    HIGH, NORMAL, LOW
  }

  public static enum Status {
    NEW,UNKNOWN,INPROGRESS,COMPLETED,FAILED
  }

  private static final String DESCRIPTION_PREFIX="GearmanJob";
  private final String DESCRIPTION;
  private byte[] handle = null;
  private String functionName = null;
  private byte[] data = null;
  private boolean backgroundJob = false;
  private JobPriority priority = null;
  private UUID uuid = null;
  private Status status = Status.NEW;
  private ByteBuffer result = ByteBuffer.allocate(Constants.GEARMAN_DEFAULT_SOCKET_RECV_SIZE);
  private ByteBuffer warning = null;
  private ByteBuffer exceptions = null;
  private GearmanClient client = null;
  private long numerator = 0;
  private long denominator = 0;
  
/**
 * Creates a new GearmanJob with the specified attributes.
 * 
 * @param functionName The name of the function to be executed by the GearmanWorker.
 * @param data The data passed to function executed by the GearmanWorker.
 * @param isBackground true if the job is to be a detached job, else false.
 * @param priority The priority to which this job is given
 */
  public GearmanJob (String functionName, byte[] data, boolean isBackground, JobPriority priority) {
    this.data = new byte[data.length];
    System.arraycopy(data, 0, this.data, 0, this.data.length);
    this.functionName = functionName;
    this.backgroundJob = isBackground;
    this.priority = priority;
    this.uuid = UUID.randomUUID();
    DESCRIPTION  = new String(DESCRIPTION_PREFIX + ":" + uuid + ":" + functionName);
  }
  
  /**
   *  Creates a new GearmanJob with the specified attributes and priority set to NORMAL.
   *  
   * @param functionName functionName The name of the function to be executed by the GearmanWorker.
   * @param data The data passed to function executed by the GearmanWorker.
   * @param isBackground true if the job is to be a detached job, else false.
   */
  public GearmanJob (String functionName, byte[] data, boolean isBackground) {
    this (functionName,data,isBackground,JobPriority.NORMAL);
  }

  /**
   * Once a job has been received by a GearmanJobServer it is a assigned a  handle which can be
   *  used by the job server to identify the job. 
   * 
   * @return the handle given to this job by the jobserver, null if the job has yet to receive
   * a handle or has not been submitted.
   */
  public byte[] getHandle() {
    return handle;
  }

  /**
   * Returns a unique ID representing this particular job.
   * 
   * @return A unique ID representing this particular job.
   */
  public byte[] getID() {
    return uuid.toString().getBytes();
  }

  /**
   * Returns the name of the function to be executed when 
   * this job is submitted.
   * 
   * @return the name of the function to be executed.
   */
  public String getFunctionName() {
    return functionName;
  }
  
  /**
   * Is this job a detached job.
   * 
   * @return True if the job is detached, otherwise false.
   */
  public boolean isBackgroundJob () {
    return backgroundJob;
  }
  
  /**
   * The priority assigned to this job.
   * 
   * @return True if the job is detached, otherwise false.
   */
  public JobPriority getPriority () {
    return priority;
  }
 
  /**
   * Returns the current status of the job.
   * @return <ul>
   * <li>{@link Status#NEW} if the job is a newly created job that has not been submitted to a JobServer</li>
   * <li>{@link Status#INPROGRESS} if the job has been submitted to a JobServer but has not completed</li>
   * <li>{@link Status#COMPLETED} if the job has been successfully completed</li>
   * <li>{@link Status#FAILED} if the job encountered an exception that required to halt execution</li>
   * </ul>
   */
  public Status getStatus (){
    return status;
  }
  
  /**
   * Jobs may send back progress updates indicating the 
   * what percentage of the job the worker believes has been
   * executed. Jobs are not required to implement this behavior
   * and as such this method should only be used if you are certain
   * the job provides these values. 
   * 
   * Note, for detached jobs, this method has no meaning, since
   * the detached job will not receive any updates from the GearmanWorker.
   *   
   * @return a denominator value that, when used in conjunction with numerator
   * value,can make up the percentage complete value. If no value has been 
   * received, this returns a default value of 0.
   */
  public synchronized long getDenominator () {
    return denominator;
  }
  
  /**
  * Jobs may send back progress updates indicating the 
   * what percentage of the job the worker believes has been
   * executed. Jobs are not required to implement this behavior
   * and as such this method should only be used if you are certain
   * the job provides these values. 
   *  
   * Note, for detached jobs, this method has no meaning, since
   * the detached job will not receive any updates from the GearmanWorker.
   *  
   * @return a numerator value that, when used in conjunction with denominator
   * value,can make up the percentage complete value.  If no value has been 
   * received, this returns a default value of 0.
   */
  public synchronized long getNumerator () {
    return numerator;
  }
  
  /**
   * Extracts any resulting data that has been received from the GearmanWorker
   * executing this job. 
   * 
   * By executing this method,the GearmanJob assumes that the
   * user no longer wishes for the GearmanJob to retain the results returned by this
   * method. As such, after executing this method the returned data is removed from 
   * the GearmanJob. Consider a scenario where a worker returns a result set consisting 
   * soley of the character 'A' and exits as completed. Upon executing this method, the
   * user will receive a byte array consisting of the character 'A'. If the user were to 
   * call this method again, the second call would result in an empty array, as the 
   * original result was cleared by the initial call. This behavior is valuable for
   * users which use results as soon as they are seen and do not require to see the
   * result set as a whole. This also has value for jobs whose results will be large
   * but who frequently return intermediate results. By clear the result set after every 
   * invocation the job will use less memory.
   * 
   * Note, for detached jobs, this method has no meaning, since
   * the detached job will not receive any updates from the GearmanWorker.
   * 
   * @return the set of results received from the GearmanWorker since the last time this
   * method was executed. If no data has been received, then return an empty array.
   */
  public synchronized byte [] extractResults() {
    byte [] res = ByteUtils.byteBufferToByteArray(result);
    result.clear();
    return res;
  }
  
  /**
   * Retrieves any resulting data that has been received from the GearmanWorker
   * executing this job. 
   * 
   * Unlike {@link #extractResults()}, this method retains the data it has received.
   * Consider the scenario where a worker sends back 3 intermediate result sets. The
   * first consisting of the character 'A', the second consisting of the character 'B' 
   * and the third consisting of the character 'C'. If we were to executing this method
   * after each result was received, the first execution would return the array {'A'}, 
   * the second would return {'A','B'} and the third would return {'A','B','C'}; 
   * 
   * Note, for detached jobs, this method has no meaning, since
   * the detached job will not receive any updates from the GearmanWorker.
   * 
   * @return the entire  set of results received from the GearmanWorker since the
   * job has commenced execution or since the last time {{@link #extractResults()} was
   * executed. If no data has been received, then return an empty array.
   */
  public synchronized byte [] getResults () {
    return ByteUtils.byteBufferToByteArray(result);
  }
 
  /**
   * Extracts any warning data that has been received from the GearmanWorker
   * executing this job and the clears the warning data from the GearmanJob. 
   * 
   * Note, for detached jobs, this method has no meaning, since
   * the detached job will not receive any updates from the GearmanWorker.
   * 
   * @return the set of warning data received from the GearmanWorker since the last 
   * time this method was executed. If no data has been received, then return an 
   * empty array.
   */
  public synchronized byte [] extractWarnings () {
    byte [] res = ByteUtils.byteBufferToByteArray(warning);
    warning.clear();
    return res;
  }
  
  /**
   * Retrieves any warning data that has been received from the GearmanWorker
   * executing this job. 
   * 
   * Note, for detached jobs, this method has no meaning, since
   * the detached job will not receive any updates from the GearmanWorker.
   * 
   * @return the entire  set of warning data received from the GearmanWorker since the
   * job has commenced execution or since the last time {{@link #extractWarnings()} was
   * executed. If no data has been received, then return an empty array.
   */
  public synchronized byte [] getWarnings () {
    return ByteUtils.byteBufferToByteArray(warning);   
  }

  /**
   * Extracts any exception data that has been received from the GearmanWorker
   * executing this job and the clears the exception data from the GearmanJob. 
   * 
   * Note, for detached jobs, this method has no meaning, since
   * the detached job will not receive any updates from the GearmanWorker.
   * 
   * @return the set of exception data received from the GearmanWorker since the last 
   * time this method was executed. If no data has been received, then return an 
   * empty array.
   */  
  public byte [] extractExceptions () {
    byte [] res = ByteUtils.byteBufferToByteArray(exceptions);
    warning.clear();
    return res;
  }
 
  /**
   * Retrieves any exception data that has been received from the GearmanWorker
   * executing this job. 
   * 
   * Note, for detached jobs, this method has no meaning, since
   * the detached job will not receive any updates from the GearmanWorker.
   * 
   * @return the entire  set of exception data received from the GearmanWorker since the
   * job has commenced execution or since the last time {{@link #extractExceptions()} was
   * executed. If no data has been received, then return an empty array.
   */
  public byte [] getExceptions () {
    return ByteUtils.byteBufferToByteArray(exceptions);
  }

  /**
   * Has the GearmanWork halted execution of this job. 
   * 
   * Note, for detached jobs, this method refers to the job submission 
   * and not the job execution itself, as detached jobs do not receive
   * updates as the job executed. 
   * 
   * @return True if the Job has received a notification from the 
   * GearmanWorker that the job has either completed or failed.
   */
  public boolean isCompleted () {
    return (status == Status.COMPLETED || status == Status.FAILED);
  }
  
  public String toString() {
    return DESCRIPTION;
  }
  
  Packet getRequest() {
    int destPos = 0;
    PacketMagic magic = PacketMagic.REQ;
    PacketType type = null;
    byte [] packetdata = null;
    byte [] fnname = ByteUtils.toAsciiBytes(functionName);
    byte [] uid = getID();
    if (priority.equals(JobPriority.HIGH)) {
      type = backgroundJob ? PacketType.SUBMIT_JOB_HIGH_BG : PacketType.SUBMIT_JOB_HIGH;
    }
    if (priority.equals(JobPriority.LOW)) {
      type = backgroundJob ? PacketType.SUBMIT_JOB_LOW_BG : PacketType.SUBMIT_JOB_LOW;
    }
    if (priority.equals(JobPriority.NORMAL)) {
      type = backgroundJob ? PacketType.SUBMIT_JOB_BG : PacketType.SUBMIT_JOB;
    }
    packetdata = new byte [fnname.length + uid.length + data.length + 2];
    System.arraycopy(fnname, 0, packetdata, destPos, fnname.length);
    destPos += fnname.length;
    packetdata[destPos++] = ByteUtils.NULL;
    System.arraycopy(uid, 0, packetdata, destPos, uid.length);
    destPos += uid.length;
    packetdata[destPos++] = ByteUtils.NULL;
    System.arraycopy(data, 0, packetdata, destPos, data.length);
    return new Packet(magic,type,packetdata);  
  }
  
  Packet getUpdateRequest () {
    PacketMagic magic = PacketMagic.REQ;
    PacketType type = PacketType.GET_STATUS;
    Packet p = new Packet(magic,type,handle);
    return p;
  }
  
  void setClient (GearmanClient cl) throws IllegalArgumentException, GearmanException {
    if (cl == null) {
      throw new IllegalArgumentException("A job's client may not be null");
    }
    
    if (this.client != null) {
      throw new GearmanException("The client for this job has already been set and may not be reset.");
    }
    this.client = cl;
  }
  
  public boolean responseComplete () {
    return isCompleted();
  }
  
  //TODO make this sync a little more fine grained and then reflect that in the setter/getters avove
  public synchronized void handleResponse(Packet event) throws GearmanException {
    PacketType pt = event.getPacketType();
    ArrayList<byte []> args = null;

    switch (pt) {
    
    case JOB_CREATED:
      args = event.getDataComponents(1);
      if (handle != null) {
        throw new GearmanException("While handling job_create for job " + handle 
            + " noticed that job objec talready has a handle " + args.get(0) + ". Duplicate create?" );
      }
      handle = args.get(0);
      //if detached, then completed
      //else inprogress and return, we return to aviod adding this to updated jobs q. 
      if (isBackgroundJob()) {
        status = Status.COMPLETED;
      }
      else {
        status = Status.INPROGRESS;
      }
      break;
      
    case WORK_STATUS:
      args = event.getDataComponents(3);
      validateJobHandle(args.get(0));
      numerator = Long.parseLong(ByteUtils.fromAsciiBytes(args.get(1)));
      denominator = Long.parseLong(ByteUtils.fromAsciiBytes(args.get(2)));
      break;
      
    case WORK_DATA:
      args = event.getDataComponents(2);
      validateJobHandle(args.get(0));
      if (result == null) {
        result = ByteBuffer.allocate(Constants.GEARMAN_DEFAULT_SOCKET_RECV_SIZE);
      }
      result = ByteUtils.addToByteBuffer(args.get(1), result);
      break;
      
    case WORK_WARNING:
      args = event.getDataComponents(2);
      validateJobHandle(args.get(0));
      if (warning == null) {
        warning = ByteBuffer.allocate(Constants.GEARMAN_DEFAULT_SOCKET_RECV_SIZE);
      }
      warning = ByteUtils.addToByteBuffer(args.get(1), warning); 
      break;
      
    case WORK_COMPLETE:
      args = event.getDataComponents(2);
      validateJobHandle(args.get(0));
      result = ByteUtils.addToByteBuffer(args.get(1), result);
      status = Status.COMPLETED;
      break;
      
    case WORK_FAIL:
      args = event.getDataComponents(1);
      validateJobHandle(args.get(0));
      status = Status.FAILED;
      break;
      
    case WORK_EXCEPTION:
      args = event.getDataComponents(2);
      validateJobHandle(args.get(0));
      if (exceptions == null) {
        exceptions = ByteBuffer.allocate(Constants.GEARMAN_DEFAULT_SOCKET_RECV_SIZE);
      }
      exceptions = ByteUtils.addToByteBuffer(args.get(1), exceptions);
      break;
      
    default:
      throw new GearmanException ("Unknown packet type " + pt);
    }
    
    client.handleJobUpdate(this);
    
  }
  
  private void validateJobHandle(byte [] rcvdHandle) throws GearmanException {
    if (!Arrays.equals(handle, rcvdHandle)) {
      throw new GearmanException ("Job handle mis-match");
    }
  }
  
}
