package gearmanij.client;

import gearmanij.Constants;
import gearmanij.GearmanException;
import gearmanij.JobImpl;
import gearmanij.Packet;
import gearmanij.PacketHeader;
import gearmanij.PacketType;
import gearmanij.client.GearmanServerRequest.RequestState;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.ArrayDeque;
import java.util.Arrays;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Queue;

// TODO HANDLE CONNECTION DROPS/IO FAILURES
// TODO cleanup closeSession()
// TODO get rid of the canRead and canWrite methods
// TODO javadoc
public class GearmanJobServerSession {

  private static final String DESCRIPTION_PREFIX="GearmanJobServerSession";
  private final String DESCRIPTION;
  private InetSocketAddress remote = null;
  private SocketChannel serverConnection = null;
  /* queue of requests that have been submitted to server and have yet to receive response */
  private Queue<GearmanServerRequest> submittedRequest = null;
  /* all request currently being handled by this session */
  private Queue<GearmanServerRequest> sessionRequests = null;
  /* all request that have  been attached to this session but have yet to be passed on the server */
  private Queue <GearmanServerRequest>newRequests  = null;
  private Deque <ByteBuffer>responsesReceived  = null;
  private Selector selector = null;
  private SelectionKey selectorKey = null; 


  GearmanJobServerSession (InetSocketAddress sa) {
    remote = sa;
    DESCRIPTION = new String(DESCRIPTION_PREFIX + ":" + Thread.currentThread().getId() + ":" + sa.toString());
  }
  
  public String toString() {
    return DESCRIPTION;
  }
  
  SocketChannel initSession () throws IOException {
    //TODO protect against duplicate init
    serverConnection = SocketChannel.open(remote);
    serverConnection.configureBlocking(false);
    serverConnection.finishConnect();
    selector = Selector.open();
    selectorKey = serverConnection.register(selector, SelectionKey.OP_WRITE | SelectionKey.OP_READ);
    responsesReceived = new ArrayDeque<ByteBuffer>();
    newRequests = new LinkedList<GearmanServerRequest>();
    sessionRequests = new LinkedList<GearmanServerRequest>();
    submittedRequest = new LinkedList<GearmanServerRequest>();
    return serverConnection;
  }
  
  InetSocketAddress getServer () {
    InetSocketAddress ret  = null;
    if (remote != null) {
      ret = new InetSocketAddress(remote.getHostName(), remote.getPort());
    }
    return ret;
  }
  
  boolean  isInitialized() {
    return (serverConnection != null && serverConnection.isConnected()); 
  }
  
  void closeSession () {
    if (serverConnection == null)
      return;
    try {
      //TODO check for NPE
      selectorKey.cancel();
      selector.close();
    } catch (IOException ioe) {
      //TODO
    } finally {
      try {
        serverConnection.close();
      } catch (IOException cioe) {
        //TODO
      }
      responsesReceived = null;
      newRequests = null;
      serverConnection = null;
    }
    //TODO figure out what to do with the session q's
  }
  
  void submitTaskToSession(GearmanServerRequest request) throws IllegalStateException {
    if (request == null) {
      throw new IllegalStateException("A null task can not be submitted to a server");
    }
    
    if (! request.getState().equals(RequestState.NEW)) {
      throw new IllegalStateException("Invalid task state: " + request.getState());
    }
    sessionRequests.add(request);
    request.setSession(this);
  }
  
  /*
   * Should be called by task when it determines that it 
   * no longer needs to talk with Server
   */
  boolean removeTaskFromSession(GearmanServerRequest request) {
    return sessionRequests.remove(request);
  }
  
  void sendTaskToServer (GearmanServerRequest request) {
    newRequests.add(request);
  }
   
  boolean canRead() throws IOException {
    selector.select();
    return selectorKey.isReadable();
  }
  
  boolean canWrite() throws IOException{
    selector.select();
    return selectorKey.isWritable();
  }
  
  void driveSessionIO () throws IOException, GearmanException {
    Iterator<GearmanServerRequest> tasks = sessionRequests.iterator();
    GearmanServerRequest curTask = null;
    while (tasks.hasNext()) {
      curTask = tasks.next();
      try {
        curTask.driveRequest();
      } finally {
        if (curTask.getState().equals(RequestState.FINISHED) || curTask.getState().equals(RequestState.EXCEPTION)) {
          tasks.remove();
        }
      }
    }
    if (canWrite())
      sendRequestToServer();
    if (canRead())
      getResponseFromServer();
    updateTaskReceivedBuffers();
  }
    
  private void sendRequestToServer () throws IOException {
    while (canWrite() && ! newRequests.isEmpty()) {
      GearmanServerRequest task =  newRequests.peek();
      ByteBuffer bb = task.getRequestAsByteBuffer();
      serverConnection.write(bb);
      if (!bb.hasRemaining()) {
        submittedRequest.add(task);
        newRequests.remove();
      }
    }
  }
  
  private void getResponseFromServer () throws IOException {
    
    //seed responsesReceived if there are none
    
    while (canRead()) {
      if (responsesReceived.isEmpty()) {
        ByteBuffer bb = ByteBuffer.allocate(Constants.GEARMAN_DEFAULT_SOCKET_RECV_SIZE);
        bb.limit(Constants.GEARMAN_PACKET_HEADER_SIZE);
        responsesReceived.add(bb);
      }
      ByteBuffer curBuf = responsesReceived.getLast();
      
      //we have read in the packet header, now figure out how big the response is and set the buffer limit
      if (curBuf.remaining() == 0 && curBuf.limit() == Constants.GEARMAN_PACKET_HEADER_SIZE) {
        byte [] header = new byte [Constants.GEARMAN_PACKET_HEADER_SIZE];
        curBuf.rewind();
        curBuf.get(header);
        PacketHeader ph = new PacketHeader(header);
        int newLimit = ph.getDataLength() + Constants.GEARMAN_PACKET_HEADER_SIZE;
        if (newLimit > curBuf.capacity()) {
          ByteBuffer bb = ByteBuffer.allocate(newLimit);
          curBuf.rewind();
          bb.put(curBuf);
          bb.position(Constants.GEARMAN_PACKET_HEADER_SIZE);
          curBuf = bb;
        } else { 
          curBuf.limit(newLimit);
        }
      } else if (curBuf.remaining() == 0) {
        //we've read the entire response, so rewind it
        curBuf.rewind();
        //now get a new buffer and add it the q
        curBuf = ByteBuffer.allocate(Constants.GEARMAN_PACKET_HEADER_SIZE);
        responsesReceived.add(curBuf);
      }
      serverConnection.read(curBuf);
    }
  }
  
  private void updateTaskReceivedBuffers() {
    Iterator<ByteBuffer> iter = responsesReceived.iterator();
    while (iter.hasNext()) {
      ByteBuffer curBuf = iter.next();
      if (curBuf.remaining() != 0)
        break;
      byte [] pb = new byte [curBuf.position()];
      curBuf.rewind();
      curBuf.get(pb);
      Packet p = new Packet(new BufferedInputStream(new ByteArrayInputStream(pb)));
      GearmanServerRequest curTask = getTaskFromPacket(p);
      curTask.addResponsePacket(p);
      iter.remove();
    }
  }
  
  private GearmanServerRequest getTaskFromPacket(Packet p) {
    GearmanServerRequest task = null;
    if (p.getPacketType().equals(PacketType.JOB_CREATED)  || 
        p.getPacketType().equals(PacketType.ECHO_RES) ||
        p.getPacketType().equals(PacketType.OPTION_RES) || 
        p.getPacketType().equals(PacketType.STATUS_RES) ||         
        p.getPacketType().equals(PacketType.ERROR)) { // TODO FATAL ERROR
      assert(!submittedRequest.isEmpty());
      task = submittedRequest.remove();
      return task;
    }
    byte[] handle = JobImpl.generateJobHandle(p.getData());
    assert (handle.length != 0);
    Iterator<GearmanServerRequest> iter = sessionRequests.iterator();
    while (iter.hasNext()) {
      GearmanServerRequest curTask = iter.next();
      if (curTask.getHandler() instanceof GearmanJob) {
        GearmanJob job = (GearmanJob) curTask.getHandler();
        if (!Arrays.equals(handle, job.getHandle()))
          continue;
        task = curTask;
        break;
      }
    }
    assert (task != null);
    return task;
  }
  
}
