package gearmanij.client;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.InetSocketAddress;
import java.util.Collection;

import gearmanij.Constants;
import gearmanij.GearmanException;

import org.junit.Ignore;
import org.junit.Test;

import junit.framework.Assert;

public class GearmanClientTest {
  
  @Test
  /*
   * Basic constructor test. Ensures that
   * a new constructed client 
   * -is active
   * -has no sessions
   */
  public void CtorTest() throws GearmanException {
    GearmanClient gc = new GearmanClient();
    Assert.assertTrue("Newly constructed client is not active.",gc.isActive());
    Assert.assertTrue("Newly constructed client has no sessions", gc.getSetOfJobServers().size() == 0);
  }
  
  @Test
  /*
   * Basic add host test. Ensures 
   * -that I can add a unique host to the client
   * -that calling hasJobServer before adding it returns false
   * -that calling hasJobServer after adding it returns true
   * -that the number of JobServers the client has is equal to the 
   *  number that has been added.
   * -verify that when we ask for the servers in the client we get them all
   */
  public void addHostsTests() throws GearmanException, IOException {
    GearmanClient gc = new GearmanClient();
    int port = Constants.GEARMAN_DEFAULT_TCP_PORT;
    int iters = 10;
    for (int i = 0; i < iters; i++) {
      Assert.assertFalse("Client claims to have jobserver localhost:" + port +  " before it was added",
          gc.hasJobServer("localhost", port + i));
      gc.addJobServer("localhost", port + i);
      Assert.assertTrue("Client claims not to have jobserver localhost:" + port +  " after it was added",
          gc.hasJobServer("localhost", port + i));
      Assert.assertTrue("Number of jobservers in client does not equal number that has been added.",
          gc.getSetOfJobServers().size() == i + 1);
    }
    Collection<InetSocketAddress> listback = gc.getSetOfJobServers();
    Assert.assertTrue("Number of jobServers returned doesn't equal number added",listback.size() == iters);
    for (int i = 0; i < iters; i++) {
      Assert.assertTrue("List of jobServers returned doesn't contain server for port " + port + i,
          listback.contains(new InetSocketAddress("localhost",port + i)));
    }
  }
  
  @Test
  /*
   * Attempt to use invalid host and port number as jobServers. All cases should
   * result in a IllegalArgumentException
   */
  public void addBadServer() throws GearmanClient.ClientShutDown, IOException{
    //invalid host
    GearmanClient gc = new GearmanClient();
    try {
      gc.addJobServer(null);
      Assert.fail("Was able to add a null job server to the client.");
    } catch (IllegalArgumentException expected) { }
    
    try {
      gc.addJobServer(null, Constants.GEARMAN_DEFAULT_TCP_PORT);
      Assert.fail("Was able to add a null job server to the client using the verbose addJobServer method.");
    } catch (IllegalArgumentException expected) { }
    
    try {
      gc.addJobServer("localhost", -1);
      Assert.fail("Was able to add a jobServer to the client using an invalid port.");
    } catch (IllegalArgumentException expected) { }
  }

  @Test
  /*
   * Basic dup add test
   * -Ensures that when we try to add a duplicate jobserver, we silently
   * ignore the add and dont increase the number of sessions
   */
  public void addDupJobServer ()  throws GearmanException, IOException {
    GearmanClient gc = new GearmanClient();
    gc.addJobServer("localhost", Constants.GEARMAN_DEFAULT_TCP_PORT);
    int numberOfServers = gc.getSetOfJobServers().size();
    gc.addJobServer("localhost", Constants.GEARMAN_DEFAULT_TCP_PORT);
    Assert.assertTrue("Adding a duplicate jobServer to a client changed the number of jobServer in the client.",
        gc.getSetOfJobServers().size() == numberOfServers); 
  }
  
  @Test
  /*
   * Basic shutdown test
   *-Ensure that I can shut down and that afterwards the client is not active 
   */
  public void basicShutdownTest() throws GearmanException, IOException {
    GearmanClient gc = new GearmanClient();
    gc.addJobServer("localhost", Constants.GEARMAN_DEFAULT_TCP_PORT);
    gc.shutdown();
    Assert.assertFalse("Client is active after attempting to shutdown client",gc.isActive());
  }
  
  @Ignore
  /*
   * Shutdown method invocation test.
   * -Ensures that after a client has been shut down, any attempt to invoke a public method on object
   *  throws a GearmanClient.ClientShutdown exception
   */
  /** TEST IS BROKEN, throws some god awful exception when trying to invoke hasJobServers, I sure problem is related ignorance of reflection */
  public void invokeShutdownTest () throws GearmanException, IOException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
    GearmanClient gc = new GearmanClient();
    gc.addJobServer("localhost", Constants.GEARMAN_DEFAULT_TCP_PORT);
    gc.shutdown();
    Method m [] = gc.getClass().getDeclaredMethods();
    for (int i = 0; i < m.length;i++) {
      if (m[i].getModifiers() != Modifier.PUBLIC || m[i].getName().equals("shutdown") || m[i].getName().equals("isActive")) {
        continue;
      }
        Object [] arglist = new Object [m[i].getGenericParameterTypes().length];
        for (int x = 0; x < arglist.length;x++) {
          arglist[x] = null;
        }
        System.out.println("Trying " + m[i].getName());
        try {
          m[i].invoke(gc, arglist);
          Assert.fail("Attempt to invoke method " + m[i] + " on a shutdown client did not result in an exception being raised");
        } catch (InvocationTargetException ite) {
          Throwable t = ite.getCause();
          Assert.assertTrue("Attempt to invoke method " + m[i] + " on a shutdown client did not result in a " + 
              GearmanClient.ClientShutDown.class + " exception being raised",t instanceof GearmanClient.ClientShutDown);
        }
          
      }
    }
  
}
